import sys

import structured_config
from structured_config import *

import utime
import unittest
from pathlib import Path


tmpdir = Path('test_structured_config_out')
if tmpdir.exists():
    def remove(entry: Path):
        if entry.is_dir():
            for _entry in entry.iterdir():
                remove(_entry)
            entry.rmdir()
        else:
            entry.unlink()
    remove(tmpdir)


ORDER_PRESERVED = True


def test_ordering():
    global ORDER_PRESERVED
    a = dict(aaa=1, bbb=2, ccc=3)
    b = dict(ccc=3, bbb=2, aaa=1)
    if list(a.keys()) == list(b.keys()):
        print("WARNING: micropython is not maintaining dict order, groups and serialise features compromised")
        ORDER_PRESERVED = False


test_ordering()


class TestStructured_config(unittest.TestCase):

    def test_nofile_structure(self):
        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        _config = Config()

        assert _config.server.url == 'https://www.example.com'
        assert _config.concurrent_connections == 32

    def test_contamination(self):
        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        config1 = Config()
        config2 = Config()

        assert config1.server.url == 'https://www.example.com'
        assert config1.concurrent_connections == 32

        assert config2.server.url == 'https://www.example.com'
        assert config2.concurrent_connections == 32

        config1.server.url = 'https://new.example.com'
        config1.concurrent_connections = 64

        assert config2.server.url == 'https://www.example.com', config2.server.url
        assert config2.concurrent_connections == 32, config2.concurrent_connections

        assert config1.server.url == 'https://new.example.com'
        assert config1.concurrent_connections == 64

    def test_structure(self):

        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_structure.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)
        _config.__save__()

        assert conffile.exists(), '_config file should exist'

        assert _config.server.url == 'https://www.example.com'
        assert _config.concurrent_connections == 32

        self.assertRaises(AttributeError, lambda: _config.missing)

    def test_structure_iter(self):
        class Config(Structure):
            concurrent_connections = 32

            def __iter__(self):
                raise ValueError("__iter__ shouldn't be called")

        conffile = tmpdir / 'test_structure_iter.json'
        _config = Config(conffile)
        _config.__save__()
        _config.__serialise__()

        assert conffile.exists(), '_config file should exist'
        assert _config.concurrent_connections == 32

        _config = Config(conffile)

    def test_equality(self):
        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        c1 = Config()
        c2 = Config()

        assert c1 == c2,  "basic equality failed"
        c1.server.username = "new"
        assert c1 != c2, "still equal after one changed"
        c2.server.username = "new"
        assert c1 == c2, "both changed equality failed"

    def test_write(self):

        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_write.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)
        _config.__save__()

        assert conffile.exists(), '_config file should exist'

        with conffile.open('r') as conffile_handle:
            orig_conf = conffile_handle.read()

        _config.concurrent_connections = 64
        _config.__save__()

        with conffile.open('r') as conffile_handle:
            post_conf = conffile_handle.read()

        assert orig_conf != post_conf

        # if fmt == CONF_PYTHON:
        #     assert 'class Config' in post_conf
        #     assert 'class server' in post_conf
        # elif fmt == CONF_JSON:
        assert '"Config":' not in post_conf
        assert '"server":' in post_conf
        # else:
        #     assert '!Config' in post_conf
        #     assert '!server' in post_conf
        assert 'url' in post_conf
        assert 'https://www.example.com' in post_conf
        assert 'username' in post_conf
        assert 'password' in post_conf
        assert 'concurrent_connections' in post_conf
        assert '64' in post_conf

    def test_round_trip(self):
        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_round_trip.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        _config.concurrent_connections = 64
        _config.server.password = 'new pass'
        _config.__save__()

        _config2 = Config(conffile)

        assert _config2.concurrent_connections == 64

        assert _config2.server == _config.server, "%s != %s" % (_config2.server, _config.server)

        assert _config2.server.url == _config.server.url
        assert _config2.server.username == _config.server.username
        assert _config2.server.password == _config.server.password

    def test_extending_structure(self):

        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_extending_structure.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        _config.concurrent_connections = 64
        _config.__save__()

        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32
            new_field = "test"

        _config2 = Config(conffile)

        assert _config2.concurrent_connections == 64

        assert _config2.server == _config.server

        assert _config2.server.url == _config.server.url
        assert _config2.server.username == _config.server.username
        assert _config2.server.password == _config.server.password
        _config2.new_field = "write"
        _config2.__save__()

        _config3 = Config(conffile)

        assert _config3.new_field == "write"

    def test_attrib_structure(self):

        class Elem(Structure):
            a = None
            b = None

        class Config(Structure):
            first = Elem()
            second = Elem(a=3, b=4)
            third = Elem

        conffile = tmpdir / 'test_attrib_structure.json'

        assert not conffile.exists(), '_config file should not already exist'
        _config = Config(conffile)

        assert _config.first.a == _config.third.a
        assert _config.second.b == 4
        _config.__save__()
        assert conffile.exists(), 'config file should exist'

        _config2 = Config(conffile)

        assert _config2.first.a == _config2.third.a
        assert _config2.second.b == 4

    def test_nested_write(self):

        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_nested_write.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        new_url = 'http://www.new.url'
        _config.server.url = new_url
        _config.__save__()

        _config2 = Config(conffile)

        assert _config2.server.url == new_url

    def test_list(self):

        class Elem(Structure):
            a = None
            b = None

            elems = List

        class Node(Structure):
            group = List(
                Elem(a=10, b=11),
                Elem(a=12, b=13),
                Elem(a=14, b=15),
            type=Elem)

        class Config(Structure):
            nodes = List(
                Node(),
            type=Node)

        conffile = tmpdir / 'test_list.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        # Check the initial settings are still as expected
        assert len(_config.nodes) == 1
        assert isinstance(_config.nodes[0], Node)

        assert len(_config.nodes[0].group) == 3
        assert isinstance(_config.nodes[0].group[0], Elem)
        assert _config.nodes[0].group[0].a == 10 and _config.nodes[0].group[0].b == 11

        # Check that the conf file reloaded matches
        _config2 = Config(conffile)
        assert _config2.nodes == _config.nodes
        _config = _config2  # type: Config
        assert isinstance(_config.nodes[0], Node)
        assert isinstance(_config.nodes[0].group[0], Elem)

        # Change a value on one of the elem and check reloaded matches
        _config.nodes[0].group[0].a = 1
        _config.__save__()

        _config3 = Config(conffile)
        assert _config3.nodes[0].group[0].a == 1

        # Change elems in the list and check reloaded matches
        _config3.nodes[0].group.pop(0)
        _config3.__save__()

        _config4 = Config(conffile)
        assert len(_config4.nodes[0].group) == 2
        assert _config4.nodes[0].group[0].a == 12
        assert _config4.nodes[0].group[1].a == 14

        assert set(dir(_config4)) == {'nodes'}
        assert isinstance(repr(_config4), str)

        assert dir(_config4.nodes[0].group)
        assert isinstance(repr(_config4.nodes[0].group), str)

        assert dir(_config4.nodes[0].group[0])
        assert isinstance(repr(_config4.nodes[0].group[0]), str)

    def test_list_contruction(self):

        class Elem(Structure):
            a = None
            b = None

            elems = List

        class Node(Structure):
            group = List(
                Elem(a=10, b=11),
                Elem(a=12, b=13),
                Elem(a=14, b=15),
                type=Elem)

        class Config(Structure):
            nodes = List(
                Node(),
                type=Node)

        _config = Config(
            nodes=[
                Node(
                    group=[
                        Elem(a=3, b=2)
                    ]
                )
            ])

        assert len(_config.nodes) == 1
        assert isinstance(_config.nodes[0], Node)
        assert len(_config.nodes[0].group) == 1
        assert isinstance(_config.nodes[0].group[0], Elem)
        assert _config.nodes[0].group[0].a == 3
        assert _config.nodes[0].group[0].b == 2

    def test_list_type_enforcement(self):
        try:
            List([1], type=float)
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "1 must be of type <class 'float'>", "unexpected error: %s" % ex

        lst = List(type=float)
        try:
            lst.append(1)
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "1 must be of type <class 'float'>", "unexpected error: %s" % ex

        class Config(Structure):
            items = List(type=int)

        config = Config()
        try:
            config.items = ["a"]
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "a must be of type <class 'int'>", "unexpected error: %s" % ex
        assert config.items == List(), "List should have been cleared. Was: %s" % config.items

        # check that errors can be ignored if the types don't match
        config._update_field(config, {"items": ["a"]}, ignore_errors=True)
        assert config.items == List(), "List should have been cleared. Was: %s" % config.items

    def test_round_trip_crc(self):
        class Config(Structure):
            class server(Structure):
                url = 'https://www.example.com'
                username = '<user>'
                password = '<password>'

            concurrent_connections = 32

        conffile = tmpdir / 'test_round_trip_crc.json'

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        assert _config.__validated__ is None, "should only be validated after written and read"

        _config.concurrent_connections = 64
        _config.server.password = 'new pass'
        _config.__save__(checksum=True)

        _config2 = Config(conffile)

        assert _config2.concurrent_connections == 64

        assert _config2.server == _config.server

        assert _config2.server.url == _config.server.url
        assert _config2.server.username == _config.server.username
        assert _config2.server.password == _config.server.password

        assert _config2.__validated__, "read in config should be crc validated"

    def test_crc_xfer(self):
        data = b'{"server": ' \
               b'{"username": "<user>", ' \
               b'"password": "<password>", ' \
               b'"url": "https://www.example.com"}, ' \
               b'"concurrent_connections": 64, "new_field": "write"}'

        repeats = 100
        start = utime.ticks_ms()

        for _ in range(repeats):
            crc = ConfigFile.crc32_xfer(data)

        data_w_crc = data + crc.to_bytes(4, 'big')
        for _ in range(repeats):
            crc_after = ConfigFile.crc32_xfer(data_w_crc)

        end = utime.ticks_ms()
        print("crc fn takes %0.2fms" % (float(end-start) / repeats))
        assert crc, "initial crc shouldn't be zero"
        assert crc_after == 0, "crc(data + crc) should be zero"

    def test_subclass(self):
        class Config(Structure):
            root_key = None
            concurrent_connections = 32

        class Server(Config):
            url = 'https://www.example.com'
            username = '<user>'
            password = '<password>'

        config1 = Server()
        assert {'root_key', 'concurrent_connections', 'url', 'username', 'password'} == set(config1.__keys__)

        assert config1.concurrent_connections == 32
        config1.concurrent_connections = 60
        assert config1.concurrent_connections == 60, config1.concurrent_connections

    def test_subclass_contamination(self):
        class Config(Structure):
            root_key = None
            concurrent_connections = 32

        class Server(Config):
            url = 'https://www.example.com'
            username = '<user>'
            password = '<password>'

        config1 = Server()
        config2 = Server()

        assert config1.url == 'https://www.example.com'
        assert config1.concurrent_connections == 32

        assert config2.url == 'https://www.example.com'
        assert config2.concurrent_connections == 32

        # modify config1
        config1.url = 'https://new.example.com'
        config1.concurrent_connections = 64

        # Check config2 wasn't contaminated
        assert config2.url == 'https://www.example.com', config2.url
        assert config2.concurrent_connections == 32, config2.concurrent_connections

        assert config1.url == 'https://new.example.com'
        assert config1.concurrent_connections == 64, config1.concurrent_connections

        # Test that new object from class isn't contaminated with changes from config1
        config3 = Server()
        assert config3 == config2, config3

    def test_wrong_structure(self):
        class Config(Structure):
            root_key = None
            concurrent_connections = 32

        class Server(Structure):
            url = 'https://www.example.com'
            username = '<user>'
            password = '<password>'

        conffile1 = tmpdir / 'test_wrong_structure_1.json'
        conffile2 = tmpdir / 'test_wrong_structure_2.json'

        assert not conffile1.exists(), '_config file should not already exist'
        assert not conffile2.exists(), '_config file should not already exist'

        config1 = Config(conffile1).__save__()
        config2 = Server(conffile2).__save__()

        assert conffile1.exists(), '_config file should exist'
        assert conffile2.exists(), '_config file should exist'

        with self.assertRaises(AttributeError):
            config3 = Config(conffile2)

    def test_obsolete_field(self):
        class OldServer(Structure):
            url = 'https://www.example.com'
            username = '<user>'
            password = '<password>'

        class Server(Structure):
            url = Deprecated()
            username = ''
            password = ''

        conffile = tmpdir / 'test_obsolete_field_1.json'

        assert not conffile.exists(), '_config file should not already exist'

        config = OldServer(conffile).__save__()

        # Check loading still works
        config2 = Server(conffile)

        try:
            _ = config2.url
            self.fail("Expected AttributeError exception not thrown")
        except AttributeError as ex:
            assert str(ex) == "url", str(ex)
        assert config2.username == '<user>'
        assert config2.password == '<password>'

        conffile2 = tmpdir / 'test_obsolete_field_2.json'
        assert not conffile2.exists(), '_config file should not already exist'

        Server(conffile2).__save__()
        contents = conffile2.read_text()
        assert 'url' not in contents

    def test_dict(self):

        class Elem(Structure):
            a = None
            b = None

            elems = Dict

        class Node(Structure):
            group = Dict({
                "m":Elem(a=10, b=11),
                "n":Elem(a=12, b=13),
                "o":Elem(a=14, b=15)
                }, type=Elem)

        class Config(Structure):
            nodes = Dict({"p":Node()}, type=Node)

        conffile = tmpdir / 'test_dict.json'
        print(conffile.absolute())

        assert not conffile.exists(), '_config file should not already exist'

        _config = Config(conffile)

        # Check the initial settings are still as expected
        assert len(_config.nodes) == 1
        assert isinstance(_config.nodes["p"], Node)

        assert len(_config.nodes["p"].group) == 3
        assert isinstance(_config.nodes["p"].group["m"], Elem)
        assert _config.nodes["p"].group["m"].a == 10 and _config.nodes["p"].group["m"].b == 11

        # Check that the conf file reloaded matches
        _config2 = Config(conffile)
        assert _config2.nodes == _config.nodes
        _config = _config2  # type: Config
        assert isinstance(_config.nodes["p"], Node)
        assert isinstance(_config.nodes["p"].group["m"], Elem)

        # Change a value on one of the elem and check reloaded matches
        _config.nodes["p"].group["m"].a = 1
        _config.__save__()

        _config3 = Config(conffile)
        assert _config3.nodes["p"].group["m"].a == 1

        # Change elems in the dict and check reloaded matches
        _config3.nodes["p"].group.pop("m")
        _config3.nodes["p"].group["z"] = Elem(a=15, b=21)
        _config3.__save__()

        _config4 = Config(conffile)
        assert len(_config4.nodes["p"].group) == 3
        assert _config4.nodes["p"].group["n"].a == 12
        assert _config4.nodes["p"].group["o"].a == 14
        assert _config4.nodes["p"].group["z"].a == 15

        assert set(dir(_config4)) == {'nodes'}
        assert isinstance(repr(_config4), str)

        assert dir(_config4.nodes["p"].group)
        assert isinstance(repr(_config4.nodes["p"].group), str)

        assert dir(_config4.nodes["p"].group["n"])
        assert isinstance(repr(_config4.nodes["p"].group["n"]), str)

    def test_dict_construction(self):

        class Elem(Structure):
            a = None
            b = None
            c = List(type=str)

            elems = Dict

        class Node(Structure):
            group = Dict({
                "m":Elem(a=10, b=11),
                "n":Elem(a=12, b=13),
                "o":Elem(a=14, b=15)
                }, type=Elem)

        class Config(Structure):
            nodes = Dict({"p":Node()}, type=Node)

        _config = Config(
            nodes={
                "z": Node(
                    group={
                        "x": Elem(a=3, b=2, c=["1","2","3"])
                    }
                )
            })

        assert len(_config.nodes) == 1
        assert isinstance(_config.nodes["z"], Node)
        assert len(_config.nodes["z"].group) == 1
        assert isinstance(_config.nodes["z"].group["x"], Elem)
        assert _config.nodes["z"].group["x"].a == 3
        assert _config.nodes["z"].group["x"].b == 2
        assert _config.nodes["z"].group["x"].c == List("1","2","3")

    def test_dict_type_enforcement(self):
        try:
            Dict({"a": 1}, type=float)
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "1 must be of type <class 'float'> (key: a)", "unexpected error: %s" % ex

        dct = Dict(type=float)
        try:
            dct["a"] = 1
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "1 must be of type <class 'float'> (key: a)", "unexpected error: %s" % ex

        class Config(Structure):
            items = Dict(type=int)

        config = Config()
        try:
            config.items["a"] = "b"
            self.fail("Expected exception due to type mismatch not thrown")
        except ValueError as ex:
            assert str(ex) == "b must be of type <class 'int'> (key: a)", "unexpected error: %s" % ex
        assert config.items == Dict(), "Dict should have been cleared. Was: %s" % config.items

        # check that errors can be ignored if the types don't match
        config._update_field(config, {"items": {"a": "b"}}, ignore_errors=True)
        assert config.items == Dict(), "Dict should have been cleared. Was: %s" % config.items

    def test_typed_fields(self):

        class Config(Structure):
            concurrent_connections = IntField(32)
            path = TypedField('12345', lambda s: "".join(reversed(s)))

        conffile1 = tmpdir / 'test_typed_fields.json'

        assert not conffile1.exists(), '_config file should not already exist'

        config1 = Config(conffile1).__save__()

        assert config1.path == '54321', config1.path

        assert '"12345"' in Path(conffile1).read_text(), Path(conffile1).read_text()

        config1.concurrent_connections = '64'

        with self.assertRaises(ValueError):
            config1.concurrent_connections = 'abc'

        assert config1.concurrent_connections == 64
        config1.__save__()

        config2 = Config(conffile1)

        assert config2.concurrent_connections == 64
        assert config2.path == '54321', config2.path

        assert dir(config2)
        assert isinstance(repr(config2.concurrent_connections), str)

    def test_selection(self):

        class NewColours(Selection):
            new_ultraviolet = 'new_ultraviolet'
            new_red = 'new_red'
            new_green = 'new_green'

        class Colours(Selection):
            ultraviolet = ''
            red = ''
            green = ''
            blue = ''
            magenta = ''
            cyan = ''
            white = ''

        class Config(Structure):
            colour1 = SelectionField('red', Colours)
            colour2 = SelectionField(Colours.blue, Colours)
            colours = MultiSelection([], Colours)
            yesno = SelectionField('yes', ['yes', 'no'])

            new_sel = SelectionField(NewColours.new_red, NewColours)

            # dep_sel = Deprecated(SelectionField('red', Colours),
            #                      new_fieldname='new_sel',
            #                      converter=lambda x: getattr(Colours, str(x)))
            #
            # dep_mul = Deprecated(MultiSelection([Colours.ultraviolet], Colours),
            #                      new_fieldname='new_sel',
            #                      converter=lambda x: getattr(Colours, str(x[0])) if x else tuple())

        conffile = tmpdir / 'test_selection.json'

        config = Config(conffile)

        assert config.colour1 == Colours.red
        assert config.colour1 == 'red'

        assert config.colour1 != 'blue'

        assert config.colour2 == Colours.blue
        assert config.colour2 == 'blue'

        with self.assertRaises(KeyError):
            config.colour2 = "none"

        assert config.yesno == 'yes'

        with self.assertRaises(KeyError):
            config.yesno = "none"

        config.colours = [Colours.red, Colours.green, 'blue']
        assert len(config.colours) == 3
        assert 'green' in config.colours
        assert Colours.blue in config.colours
        with self.assertRaises(KeyError):
            config.colours = [Colours.red, 'dead']

        with self.assertRaises((KeyError, AttributeError)):
            config.colours.append('dead')

        # Reload from file
        config = Config(conffile)

        config.colours = [Colours.red, Colours.green, 'blue']
        assert len(config.colours) == 3
        assert 'green' in config.colours
        assert Colours.blue in config.colours

    def test_doc(self):

        class Config(Structure):
            zero = Field(0)                              | "zero"
            nope = BoolField('no')                       | "should be false"
            three = SelectionField('3', ['1', '2', '3']) | "three"

        conffile = tmpdir / 'test_doc.json'
        config = Config(conffile)

        assert config.zero == 0
        assert 'zero' in config.__fdoc__('zero')

        assert 'false' in config.__fdoc__('nope')
        assert config.nope is False

        assert 'three' in config.__fdoc__('three')
        assert config.three == '3'
        config.three = '1'
        config.__save__()
        del config

        config2 = Config(conffile)

        assert config2.zero == 0
        assert 'zero' in config2.__fdoc__('zero')

        assert 'false' in config2.__fdoc__('nope')
        assert config2.nope is False

        assert 'three' in config2.__fdoc__('three')
        assert config2.three == '1'

    def test_serialise(self):

        class Colours(Selection):
            ultraviolet = ''
            red = ''
            green = ''
            blue = ''
            magenta = ''
            cyan = ''
            white = ''

        class Elem(Structure):
            a = None
            b = None

        class Node(Structure):
            a_list = List(
                Elem(a=10, b=11),
                Elem(a=12, b=13),
                Elem(a=14, b=15),
            type=Elem)

            a_dict = Dict({
                "m":Elem(a=10, b=11),
                "n":Elem(a=12, b=13),
                "o":Elem(a=14, b=15)
                }, type=Elem)

        class Config(Structure):
            nodes = List(
                Node(),
            type=Node)

            zero = Field(0)                              | "zero"
            nope = BoolField('no')                       | "should be false"
            three = SelectionField('3', ['1', '2', '3']) | "three"

            class Inner(Structure):
                concurrent_connections = IntField(32)
                rf = RangedFloat(10, 1.0, 19.5)

            colour1 = SelectionField('red', Colours)

            deadtome = Deprecated()

        conffile1 = tmpdir / 'test_serialise_1.json'
        assert not conffile1.exists(), '_config file should not already exist'
        config1 = Config(conffile1).__save__()

        with self.assertRaises(KeyError, msg="'not-a-color' is not an allowed value"):
            config1.colour1 = "not-a-color"

        ser = config1.__serialise__()
        print('\n', ser)
        Config2 = structured_config.__deserialise__(ser)
        conffile2 = tmpdir / 'test_serialise_2.json'

        config2 = Config2(conffile2).__save__()
        ser2 = config2.__serialise__()
        print('\n', config2)

        self.assertEqual(config2.__fdoc__("zero"), "zero")
        # ensure deserialised object has the same limit behaviours
        with self.assertRaises(KeyError, msg="'not-a-color' is not an allowed value"):
            config1.colour1 = "not-a-color"

        assert config1 == config2, "Objects are not the same"
        # TODO check SelectionField behaviour survives

        if not ORDER_PRESERVED:
            print("SKIPPED checking file and ser obj equality: dict order not preserved")
        else:
            assert ser == ser2, "Serialised objects are not the same"
            assert conffile1.read_bytes() == conffile2.read_bytes(), "Files are not the same"

    def test_ctor_value_type_enforcement(self):
        class Config(Structure):
            class field(Structure):
                value = "some string"
                list_field = List(type=int)
                dict_field = Dict(type=bool)

        class Field2(Structure):
            value2 = "another string"

        with self.assertRaises(TypeError, msg="conf be either a matching type or a dict: <class 'Field2'>"):
            Config(field=Field2())

        with self.assertRaises(TypeError, msg="conf be either a matching type or a dict: <class 'Field2'>") as cm:
            Config().field = Field2()

        config = Config(field={"value": "a"})
        assert config.field.value == "a", "config.field.value was %s" % config.field.value

        config = Config(field={"list_field": [1]})
        assert config.field.list_field == List([1]), "config.field.list_field was %s" % config.field.list_field

        config = Config(field={"list_field": List([1], type=int)})
        assert config.field.list_field == List([1]), "config.field.list_field was %s" % config.field.list_field

        with self.assertRaises(ValueError, msg="1 must be of type <class 'bool'> (key: key)"):
            Config(field={"dict_field": {"key": 1}})

        config = Config(field={"dict_field": {"key": True}})
        assert config.field.dict_field == Dict({"key": True}), "config.field.dict_field was %s" % config.field.dict_field

    def test_deprecated_field_update(self):
        class Config(Structure):
            new_field = IntField(1)
            old_field = Deprecated(IntField(5), "new_field", lambda data: -1 * data)
            dead_field = Deprecated()

        config = Config()

        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == 1, 'deprecated field shouldnt override initial value on new_field'

        config.dead_field = 12
        self.assertRaises(AttributeError, lambda: config.dead_field)

        config.old_field = "4"
        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == -4, 'new_field should be set with value from deprecated field'

        config.old_field = 8
        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == -8, 'new_field should be updated with value from deprecated field'

        new_config = Config(dead_field="aaa")
        self.assertRaises(AttributeError, lambda: new_config.dead_field)
        assert new_config.new_field == 1, 'new_field should not be affected'

        new_config = Config(old_field=5)
        assert new_config.new_field == -5, 'new_field was %s after direct update from old_field'%new_config.new_field

        new_config = Config(new_field=10)
        assert new_config.new_field == 10, 'new_field was %s after direct update from new_field'%new_config.new_field

        new_config = Config(old_field=5, new_field=10)
        assert new_config.new_field == -5, 'new_field was %s after direct update with both fields'%new_config.new_field

        new_config = Config(new_field=10, old_field=5)
        assert new_config.new_field == -5, 'new_field was %s after direct update with both fields'%new_config.new_field

    def test_deprecated_field_collections(self):
        class Config(Structure):
            new_field = List(1, type=int)
            old_field = Deprecated(List(type=int), "new_field", lambda data: [-1 * v for v in data])

        config = Config()
        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == List(1), 'deprecated field shouldnt override initial value on new_field'

        config.old_field = [4]
        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == List(-4), 'new_field should be set with value from deprecated field: %s' % config.new_field

        config.old_field = [4, -10]
        self.assertRaises(AttributeError, lambda: config.old_field)
        assert config.new_field == List(-4, 10), 'new_field should be set with value from deprecated field: %s' % config.new_field

        new_config = Config(old_field=[5])
        assert new_config.new_field == List(-5), 'new_field was %s after direct update from old_field'%new_config.new_field

        new_config = Config(new_field=[10])
        assert new_config.new_field == List(10), 'new_field was %s after direct update from new_field'%new_config.new_field

        new_config = Config(old_field=[5], new_field=[10])
        assert new_config.new_field == List(-5), 'new_field was %s after direct update with both fields'%new_config.new_field

        new_config = Config(new_field=[10], old_field=[5])
        assert new_config.new_field == List(-5), 'new_field was %s after direct update with both fields'%new_config.new_field

        new_config.new_field = None
        assert new_config.new_field == List(), 'new_field was %s after direct update with both fields'%new_config.new_field

    def test_deprecated_field_read(self):
        def conv(data):
            return -1 * data if data is not None else None

        def conv_list(data):
            return [-1 * v for v in data]

        def conv_dict(data):
            return { k:-1 * v for k, v in data.items() }

        class ConfigOnDisk(Structure):
            old_field = IntField(5)
            old_list = List(1,2,3,type=int)
            old_dict = Dict({"a": 1, "b": 2},type=int)

        class NewConfig(Structure):
            new_field = IntField(91)
            old_field = Deprecated(IntField(95), "new_field", conv)

            new_list = List(4,5,6,type=int)
            old_list = Deprecated(List(1,2,3,type=int), "new_list", conv_list)

            new_dict = Dict({"c": 3, "d": 4},type=int)
            old_dict = Deprecated(Dict({"a": 5, "b": 6},type=int), "new_dict", conv_dict)

        class NewConfig2(Structure): # same as NewConfig but with field order reversed
            old_field = Deprecated(IntField(95), "new_field", conv)
            new_field = IntField(91)

            old_list = Deprecated(List(1,2,3,type=int), "new_list", conv_list)
            new_list = List(7,8,9,type=int)

            old_dict = Deprecated(Dict({"a": 5, "b": 6},type=int), "new_dict", conv_dict)
            new_dict = Dict({"c": 3, "d": 4},type=int)

        conffile = tmpdir / 'test_deprecated_field_read.json'

        ConfigOnDisk(conffile).__save__()
        print(conffile.read_text())

        reloaded = NewConfig(conffile)
        assert reloaded.new_field == -5, 'new_field was %s' % reloaded.new_field
        self.assertRaises(AttributeError, lambda: reloaded.old_field)

        assert reloaded.new_list == List(-1,-2,-3), 'new_list was %s' % reloaded.new_list
        self.assertRaises(AttributeError, lambda: reloaded.old_list)

        assert reloaded.new_dict == Dict({"a": -1, "b": -2}), 'new_dict was %s' % reloaded.new_dict
        self.assertRaises(AttributeError, lambda: reloaded.old_dict)

        reloaded = NewConfig2(conffile)
        assert reloaded.new_field == -5, 'new_field was %s' % reloaded.new_field
        self.assertRaises(AttributeError, lambda: reloaded.old_field)

        assert reloaded.new_list == List(-1,-2,-3), 'new_list was %s' % reloaded.new_list
        self.assertRaises(AttributeError, lambda: reloaded.old_list)

        assert reloaded.new_dict == Dict({"a": -1, "b": -2}), 'new_dict was %s' % reloaded.new_dict
        self.assertRaises(AttributeError, lambda: reloaded.old_dict)

        conffile.unlink()

        new_config = NewConfig(conffile)
        new_config.new_field = 10
        assert new_config.new_field == 10, 'new_field was %s' % new_config.new_field

        new_config.new_list = [20, 21, 22]
        assert new_config.new_list == List(20, 21, 22), 'new_list was %s' % new_config.new_list

        new_config.new_dict = {"c": 1, "d": 2}
        assert new_config.new_dict == Dict({"c": 1, "d": 2}), 'new_dict was %s' % new_config.new_dict

        new_config.__save__()

        reloaded = NewConfig2(conffile)
        assert reloaded.new_field == 10, 'new_field was %s after reload' % reloaded.new_field
        assert reloaded.new_list == List(20, 21, 22), 'new_list was %s' % reloaded.new_list
        assert reloaded.new_dict == Dict({"c": 1, "d": 2}), 'new_dict was %s' % reloaded.new_dict

    def test_deprecated_field_nested_with_migrations(self):
        class Config(Structure):
            new_field = StrField("c")
            teenage_field = Deprecated(StrField("b"), "new_field", lambda data: data + "|b")
            old_field = Deprecated(StrField("a"), "teenage_field", lambda data: data + "|a")

        config = Config()
        assert config.new_field == 'c', 'deprecated field shouldnt override initial value: %s' % config.new_field

        config.old_field = "x"
        assert config.new_field == 'x|a|b', 'some migrations missing: %s' % config.new_field

        config.teenage_field = "x"
        assert config.new_field == 'x|b', 'some migrations missing: %s' % config.new_field

        config.new_field = "x"
        assert config.new_field == 'x', 'unexpected migrations: %s' % config.new_field

    def test_groups(self):
        if not ORDER_PRESERVED:
            print("SKIPPED: dict order not preserved which breaks groups")
            return

        class Config(Structure):
            group_1 = Group("group 1")
            string_field = StrField("a")
            list_field = List()

            group_2 = Group("group 2")
            dict_field = Dict()
            class subtree(Structure):
                url = ""

        config = Config()
        self.assertEqual(config.__fgroup__("string_field"), "group 1")
        self.assertEqual(config.__fgroup__("list_field"), "group 1")
        self.assertEqual(config.__fgroup__("dict_field"), "group 2")
        self.assertEqual(config.__fgroup__("subtree"), "group 2")

        ser = config.__serialise__()
        Config2 = structured_config.__deserialise__(ser)
        config2 = Config2()
        self.assertEqual(config2.__fgroup__("string_field"), "group 1")
        self.assertEqual(config2.__fgroup__("list_field"), "group 1")
        self.assertEqual(config2.__fgroup__("dict_field"), "group 2")
        self.assertEqual(config2.__fgroup__("subtree"), "group 2")

    def test_serialise_list(self):
        class Config(Structure):
            interval = (
                List([1.0, 1.5, 2.0], type=float) | "The interval between reads"
            )

        config = Config()
        ser = config.__serialise__()
        print(ser)
        Config2 = structured_config.__deserialise__(ser)
        config2 = Config2()

        assert config2 == config


if __name__ == '__main__':
    unittest.main()



# Micropython structured_config based on https://gitlab.com/alelec/structured_config
import errno
import ujson
import copy
from pathlib import Path
from collections import OrderedDict

try:
    import copy
    copy.dispatch_table
except AttributeError:
    copy.dispatch_table = {}

_deepcopy_dispatch = getattr(copy, "_deepcopy_dispatch")
if OrderedDict not in _deepcopy_dispatch:
    _deepcopy_dispatch[OrderedDict] = _deepcopy_dispatch[dict]

from copy import copy, deepcopy

OMIT_VALUE = object()

class Structure:

    def __new__(cls, *args, **kwargs):
        __template = "__" + cls.__name__
        if not hasattr(cls, __template):
            # Copy the attributes from the class def to a new
            # template dict so we can delete the (shared) class
            # attributes
            __data__ = OrderedDict()

            if cls is not Structure and Structure not in cls.__bases__:
                # Ensure base classes have been correctly templated first
                for c in cls.__bases__:
                    if issubclass(c, Structure):
                        __data__.update(c().__template__())

            if hasattr(cls, '__dict__'):
                # preserves proper field ordering in micropython
                keys = cls.__dict__.keys()
            else:
                keys = dir(cls)
            for key in keys:
                if key.startswith('_'):
                    continue
                value = getattr(cls, key)
                __data__[key] = _instantiated(value)
                delattr(cls, key)
            setattr(cls, __template, __data__)
            for field in __data__.values():
                if hasattr(field, '_fields_registered'):
                    field._fields_registered(__data__)

        # Create instance
        self = object.__new__(cls)
        _inst = OrderedDict()

        # Ensure instance has copy of attributes and defaults from definition
        template = getattr(cls, __template)
        for key, value in template.items():
            field_initialiser = getattr(value, "_field_initialiser", lambda: deepcopy(value))
            value = field_initialiser()
            if value != OMIT_VALUE:
                _inst[key] = value

        self.__keys__ = _inst.keys()
        self.__group__ = ""
        self._configfile = None
        self._readonly = False
        self._inst = _inst
        self._isset = set()
        return self

    @classmethod
    def __template__(cls):
        return getattr(cls, "__" + cls.__name__)

    def __call__(self, *args, **kwargs):
        return self

    def __init__(self, *args, **kwargs):
        if args:
            if not isinstance(args[0], (str, Path)):
                raise ValueError("%s: arg[0] should be config file, not %s" % (self.__class__.__name__, args[0]))
            self._configfile = ConfigFile(args[0], self)

        for key, val in kwargs.items():
            self[key] = val

    def __items__(self):
        for key in self.__keys__:
            if key.startswith('_'):
                continue
            yield key, getattr(self, key)

    def __len__(self):
        return len(self._inst)

    __iter__ = __items__

    def __contains__(self, item):
        return item in self.__keys__

    def __setattr__(self, key, value):
        """
        This function is replaced by __setitem__ once the structure object is created
        """
        object.__setattr__(self, key, value)
        if key == '_inst':
            object.__setattr__(self, '__setattr__', self.__setitem__)

    def __setitem__(self, key, value):
        if self._readonly:
            raise AttributeError("Cannot change values of readonly", self.__class__.__name__)
        try:
            field = self.__template__()[key]
            self._isset.add(key)
            update_field = getattr(field, '_update_field', None)
            if update_field:
                update_field(self._inst.get(key, None), value, parent=self)
            else:
                convert_value = getattr(field or value, 'converter', None)
                if convert_value:
                    value = convert_value(value)
                if value is not OMIT_VALUE:
                    self._inst[key] = value
                else:
                    self._isset.remove(key)
                    del self._inst[key]
        except KeyError:
            if key.startswith("_"):
                object.__setattr__(self, key, value)
            else:
                raise

    def __dir__(self):
        return self.__keys__

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError as ex:
            raise AttributeError(ex.args[0])

    def __getitem__(self, item):
        return self._inst[item]

    def __repr__(self):
        return "<%s:{%s}>" % (self.__class__.__name__, ', '.join(("%s:%s" % i for i in self)))

    def __eq__(self, other):
        if type(other).__name__ == type(self).__name__:
            for key, ival in self._inst.items():
                try:
                    oval = other[key]
                except KeyError:
                    return False
                if ival != oval:
                    return False
            return True

    def __deepcopy__(self, memo):
        n = self.__class__(
            **deepcopy(
                {key: self._inst[key] for key in self.__keys__}, memo
            )
        )
        n.__group__ = self.__group__
        return n

    def __save__(self, checksum=False):
        conf = self._configfile
        if not conf:
            raise ValueError("No config file set")
        conf.dump(checksum=checksum)
        return self

    @property
    def __validated__(self):
        if self._configfile:
            return self._configfile.validated
        return None

    def __fgroup__(self, field):
        """
        Returns the __group__ for the given field
        :param str field: structure field to get doc for
        :return: str
        """
        return self.__template__().get(field).__group__

    def __fdoc__(self, field=None):
        """
        Returns the __doc__ for the given field
        :param str field: structure field to get doc for
        :return: str
        """
        return self.__template__().get(field).__doc__

    def __serialise__(self, raw=False):
        ret = OrderedDict()
        for key, val in self._inst.items():
            field = self.__template__()[key]
            if hasattr(val, '__serialise__'):
                val = val.__serialise__(raw=True)
            elif hasattr(field, '__serialise__'):
                if isinstance(field, Field):
                    # TODO fields should have static serialise accessor
                    new_val = deepcopy(field)
                    new_val.value = val
                    val = new_val
                val = val.__serialise__(raw=True)
            else:
                val = OrderedDict(type=type(val).__name__, value=[val])
            ret[key] = val

        ret = OrderedDict(
            type="Structure",
            name=self.__class__.__name__,
            fields=ret
        )
        if self.__group__:
            ret["group"] = self.__group__
        return ret if raw else ujson.dumps(ret)

    def _field_initialiser(self):
        return deepcopy(self)

    @staticmethod
    def _update_field(_structure, conf, parent=None, ignore_errors=False):
        if conf is None:
            return

        if type(conf) not in (type(_structure), dict, OrderedDict):
            raise TypeError("conf be either a matching type or a dict: %s" % type(conf))

        warn = []
        set_keys = 0
        num_keys = len(conf)
        _structure._isset = set()
        for key, field in _structure.__template__().items():
            if key not in _structure and key not in conf:
                # A deprecated field not have a value in either structure or incoming
                continue

            if key not in conf and key in _structure._isset:
                print("Skipping key '%s': absent in incoming structure and has already been set" % key)
                continue

            update_field = getattr(field, '_update_field', None)
            if update_field:
                update_field(_structure._inst.get(key, None), conf.get(key, None),
                                parent=_structure, ignore_errors=ignore_errors)
                set_keys += 1 if key in conf else 0
            else:
                if key in conf:
                    _structure[key] = conf[key]
                    set_keys += 1
                else:
                    warn.append("WARN: Config file does not contain %s, default will be used" % key)

        if not ignore_errors:
            extra = []
            if set_keys != num_keys:
                # There are extra keys in the config file, it's probably the incorrect structure
                for key in conf:
                    if key not in _structure:
                        extra.append(key)

                raise AttributeError("Config file has extra keys: %s" % extra)

            for msg in warn:
                print(msg)

    @staticmethod
    def _unwrap_value(structure):
        fields = structure.__template__()
        return { key: maybe_unwrap_value(value, fields[key]) for key, value in structure._inst.items() }


def _instantiated(cls):
    if type(cls) == type and issubclass(cls, (Structure, List, Dict)):
        return cls()
    return cls


def maybe_unwrap_value(val, field=None):
    unwrap_field = getattr(field or val, '_unwrap_value', None)
    if unwrap_field:
        val = unwrap_field(val)
    elif isinstance(val, list):
        val = List._unwrap_value(val)
    elif isinstance(val, dict):
        val = Dict._unwrap_value(val)
    return val

def __deserialise__(ser):
    def _issubclass(cls, t):
        try:
            return issubclass(cls, t)
        except TypeError:
            return False

    types = OrderedDict(str=str, int=int, float=float, bool=bool, Field=Field, NoneType= lambda *_:None, **{
        k: t for k, t in globals().items() if _issubclass(t, Field)
    })

    if isinstance(ser, str):
        ser = ujson.loads(ser)

    def new_structure(defs):
        fields = OrderedDict()
        for k, v in defs['fields'].items():
            fields[k] = decode(v)
        return type(defs['name'], (Structure,), fields)

    def decode(item):
        if item['type'] == 'Structure':
            t = new_structure(item)
            cls = types.get(item['name'], t)
            types[item['name']] = cls
            obj = cls(**OrderedDict(t().__items__()))
        elif item['type'] == 'List':
            obj = List(*(decode(v) for v in item['value']), type=decode(item['element_type']).__class__)
        elif item['type'] == 'Dict':
            obj = Dict(type=decode(item['element_type']).__class__)
            obj.update(OrderedDict(((k, decode(v)) for k, v in item['value'].items())))
        else:
            obj = types[item['type']](*item['value'])

        if hasattr(obj, "__doc__"):
            obj.__doc__ = item.get("doc", "")

        if hasattr(obj, "__group__"):
            obj.__group__ = item.get("group", "")
        return obj

    return decode(ser).__class__


class List(list):
    """
    Overridden list to allow us to wrap functions for automatic write.
    This is required as we can't wrap/replace the builtin list functions
    """

    def __init__(self, *args, type=None, **kwargs):
        self.type = type
        self.__doc__ = ""
        self.__group__ = ""
        items = []
        if len(args) == 1 and isinstance(args[0], (list, tuple, set, )):  # todo: GeneratorType
            args = args[0]
        super(List, self).__init__(items, **kwargs)
        self._update_field(self, args)

    def append(self, elem):
        if self.type and not isinstance(elem, self.type):
            raise ValueError("%s must be of type %s" % (elem, self.type))
        super(List, self).append(elem)

    def extend(self, elems):
        if self.type and any((not isinstance(elem, self.type) for elem in elems)):
            raise ValueError("%s must all be of type %s" % (elems, self.type))
        super(List, self).extend(elems)

    def insert(self, index, obj):
        if self.type and not isinstance(obj, self.type):
            raise ValueError("%s must be of type %s" % (obj, self.type))
        super(List, self).insert(index, obj)

    def __eq__(self, other):
        if type(other).__name__ == self.__class__.__name__:
            return list(self) == list(other)

    def __deepcopy__(self, memo):
        n = List(*deepcopy(list(self), memo), type=self.type)
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def __or__(self, other):
        self.__doc__ = other
        return self

    def __serialise__(self, raw=False):
        def encode(val):
            if hasattr(val, '__serialise__'):
                val = _instantiated(val)
                return val.__serialise__(raw=True)
            elif type(val) is type:
                return OrderedDict(type=val.__name__, value=[])
            else:
                return OrderedDict(type=type(val).__name__, value=[val])

        obj = OrderedDict(
            type=self.__class__.__name__,
            value=[encode(v) for v in self],
            element_type=encode(self.type),
            doc=self.__doc__
        )
        if self.__group__:
            obj["group"] = self.__group__
        return obj if raw else ujson.dumps(obj)

    @staticmethod
    def _update_field(_list, conf, parent=None, ignore_errors=False):
        if conf is not None and type(conf) not in (type(_list), list, set, tuple):
            raise TypeError("conf be either a matching type or a list: %s" % type(conf))

        _list.clear()
        if conf is None:
            return
        conv = _list.type
        for item in conf:
            if not conv:
                _list.append(item)
            else:
                if isinstance(item, dict):
                    _list.append(conv(**item))
                elif isinstance(item, (list, tuple)):
                    _list.append(conv(*item))
                else:
                    try:
                        if isinstance(item, conv):
                            _list.append(item)
                        elif not ignore_errors:
                            raise ValueError("%s must be of type %s" % (item, conv))
                    except TypeError:
                        _list.append(conv(item))

    @staticmethod
    def _unwrap_value(val):
        return [maybe_unwrap_value(v) for v in val]


class Dict(OrderedDict):
    """
    Overridden dict to allow us to wrap functions for automatic write.
    Wrapping the builtins the same way as List didn't work, __setitem__
    would not fire the config writer
    """

    def __init__(self, *args, type=None):
        self.type = type
        self.__doc__ = ""
        self.__group__ = ""
        super(Dict, self).__init__(*args)
        if self.type:
            for key, value in self.items():
                if not isinstance(value, self.type):
                    raise ValueError("%s must be of type %s (key: %s)" % (value, self.type, key))

    def __setitem__(self, key, value):
        if self.type and not isinstance(value, self.type):
            raise ValueError("%s must be of type %s (key: %s)" % (value, self.type, key))
        return super(Dict, self).__setitem__(key, value)

    def __eq__(self, other):
        if type(other).__name__ == self.__class__.__name__:
            return dict(self.items()) == dict(other.items())

    def __deepcopy__(self, memo):
        n = Dict(deepcopy(OrderedDict(self.items()), memo), type=self.type)
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def __or__(self, other):
        self.__doc__ = other
        return self

    def __serialise__(self, raw=False):
        def encode(val):
            if hasattr(val, '__serialise__'):
                val = _instantiated(val)
                return val.__serialise__(raw=True)
            elif type(val) is type:
                return OrderedDict(type=val.__name__, value=[])
            else:
                return OrderedDict(type=type(val).__name__, value=[val])

        obj = OrderedDict(
            type=self.__class__.__name__,
            value={k: encode(v) for k, v in self.items()},
            element_type=encode(self.type),
            doc=self.__doc__
        )
        if self.__group__:
            obj["group"] = self.__group__
        return obj if raw else ujson.dumps(obj)

    @staticmethod
    def _update_field(_dict, conf, parent=None, ignore_errors=False):
        if conf is None:
            _dict.clear()
            return

        if type(conf) not in (type(_dict), dict, OrderedDict):
            raise TypeError("conf be either a matching type or a dict: %s" % type(conf))

        conv = _dict.type
        for key, value in conf.items():
            existing = _dict.get(key, None)

            if conv:
                if isinstance(value, dict):
                    _dict[key] = conv(**value)
                elif isinstance(value, (list, tuple)):
                    _dict[key] = conv(*value)
                elif isinstance(value, conv):
                    _dict[key] = value
                elif not ignore_errors:
                    raise ValueError("%s must be of type %s (key: %s)" % (value, conv, key))
            elif existing is not None:
                update_field = getattr(existing, '_update_field', None)
                if update_field:
                    update_field(existing, value, parent=_dict, ignore_errors=ignore_errors)
                else:
                    _dict[key] = value
            else:
                _dict[key] = value

        for key in list(_dict.keys()):
            if key not in conf:
                del _dict[key]

    @staticmethod
    def _unwrap_value(structure):
        return {key: maybe_unwrap_value(val) for key, val in structure.items()}


class ConfigFile:
    def __init__(self, configfile, structure, load=True, validate='auto'):
        self._configfile = Path(configfile)
        self._structure = structure
        self.validated = None
        if load:
            try:
                self.load(validate)
            except OSError as ex:
                if ex.args[0] != errno.ENOENT:
                    raise

    def load(self, validate='auto', ignore_errors=False):
        """
        Load config from pre-configured file
        :param str|bool validate: set to true to force validation, false to ignore, auto otherwise
        :return:
        """
        with self._configfile.open('r') as fh:
            data = fh.read()
            lastline = data.split('\n')[-1]
            if validate is True or (lastline.startswith('#') and 'crc' in lastline.lower()):
                if self.crc32_xfer(memoryview(data)) != 0:
                    self.validated = False
                    if validate:
                        raise ValueError("Checksum Failure")
                # Strip crc line off end
                data = data[0:-len(lastline)]
                self.validated = True
            try:
                conf = ujson.loads(data)
            except:
                print("Invalid json data could not be loaded:", data)
                raise

        Structure._update_field(self._structure, conf, ignore_errors=ignore_errors)

    def dump(self, checksum=False):
        _dict = Structure._unwrap_value(self._structure)
        # Pre-convert to json to avoid deleting file if there are errors
        _json = ujson.dumps(_dict)
        if checksum:
            _json += "\n# CRC-32/XFER: "
            _json = bytearray(_json)
            # crc = ubinascii.crc32(_json, 0x000000AF)
            crc = self.crc32_xfer(_json)

            # Why we find 0x0a (new line):
            # Because we use eol to find length of CRC,
            # if CRC itself has new line we fail to determine length of CRC
            while 0x0a in crc.to_bytes(4, 'big'):
                _json.extend(b' ')
                crc = self.crc32_xfer(_json)

            _json.extend(crc.to_bytes(4, 'big'))
        if not self._configfile.parent.exists():
            self._configfile.parent.mkdir(parents=True)
        with self._configfile.open('w') as fh:
            fh.write(_json)

    @staticmethod
    def crc32_xfer(data):
        """
        Returns the 32 bit CRC (XFER algorithm) of the provided bytes

        Compatible with CRC-32/XFER from https://crccalc.com/?method=crc32
        If the returned crc is appended to the original data and run back through
        this function the result will be 0.
        :param bytes,bytearray data: bytes to calculate crc on
        :return int: crc of input data
        """
        _poly = 0x000000AF
        crc = 0x00000000
        highbit = 1 << (32 - 1)
        mask = ((highbit - 1) << 1) | 0x1
        poly = _poly
        shift = 32 - 8
        diff8 = -shift
        if diff8 > 0:
            mask = 0xFF
            crc <<= diff8
            shift = 0
            highbit = 0x80
            poly = _poly << diff8

        for byte in data:
            crc ^= (byte << shift)
            for i in range(8):
                if crc & highbit:
                    crc = (crc << 1) ^ poly
                else:
                    crc = (crc << 1)
            crc &= mask
        if diff8 > 0:
            crc >>= diff8
        return crc


##########
# Fields #
##########


class Field(object):
    """
    Base class for fields that allows capturing of comments or'ed with the field
    """
    def __init__(self, value):
        self.value = value
        self.__doc__ = ''
        self.__group__ = ''

    def __or__(self, other):
        self.__doc__ = other
        return self

    def __getstate__(self) -> tuple:
        return self.value,

    def __setstate__(self, state: tuple):
        self.value, = state

    def __deepcopy__(self, memodict=None):
        n = self.__class__(deepcopy(self.value, memodict or {}))
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def __str__(self):
        return "<%s>(%s)|%s" % (self.__class__.__name__, ",".join((repr(a) for a in self.__getstate__())), self.__doc__)

    def __serialise__(self, raw=False):
        ret = OrderedDict(
            type=self.__class__.__name__,
            value=self.__getstate__(),
            doc=self.__doc__
        )
        if self.__group__:
            ret["group"] = self.__group__
        return ret if raw else ujson.dumps(ret)

    def _field_initialiser(self):
        return self.value


class TypedField(Field):
    def __init__(self, value, converter, writer=None, **kwargs):
        self.converter = converter
        self.writer = writer if writer else converter
        super(TypedField, self).__init__(value=converter(value))

        # cls = self.__class__
        # self._yaml_tag = '!' + cls.__name__
        #
        # yaml.add_constructor(self._yaml_tag, self._from_yaml)

    def __deepcopy__(self, memodict=None):
        if self.__class__ is TypedField:
            n = self.__class__(deepcopy(self.value, memodict or {}), self.converter, self.writer)
        else:
            n = self.__class__(deepcopy(self.value, memodict or {}))
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def _unwrap_value(self, val):
        return self.writer(val)

    # @classmethod
    # def _from_yaml(cls, loader, node):
    #     return loader.construct_yaml_object(node, cls)


class IntField(TypedField):
    def __init__(self, value):
        super(IntField, self).__init__(value, lambda v: None if v is None else int(v))


class StrField(TypedField):
    def __init__(self, value):
        super(StrField, self).__init__(value, lambda v: None if v is None else str(v))


class FloatField(TypedField):
    def __init__(self, value):
        super(FloatField, self).__init__(value, self.__check__)

    @staticmethod
    def __check__(value):
        if value is None:
            return None
        if (value * 0) != 0:  # test for nan/inf:
            return "nan"
        return float(value)


class PathField(TypedField):
    def __init__(self, value):
        super(PathField, self).__init__(value, Path)


class RangedNumber(TypedField):
    def __init__(self, value, min, max):
        self.min = min
        self.max = max
        super(RangedNumber, self).__init__(value, self.__check__)

    def __check__(self, value):
        if self.min <= value <= self.max:
            return value
        raise ValueError("%s out of range (%s - %s)" % (value, self.min, self.max))

    def update_range(self, min, max):
        self.min = min
        self.max = max

    def __deepcopy__(self, memodict=None):
        n = self.__class__(*deepcopy((self.value, self.min, self.max), memodict or {}))
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def __getstate__(self):
        return self.value, self.min, self.max

    def __setstate__(self, *args):
        self.value, self.min, self.max = args


class RangedFloat(RangedNumber):
    def __check__(self, value):
        try:
            return super(RangedFloat, self).__check__(float(value))
        except TypeError:
            raise ValueError("%s out of range (%s - %s)" % (value, self.min, self.max))


class RangedInt(RangedNumber):
    def __check__(self, value):
        try:
            return super(RangedInt, self).__check__(int(value))
        except TypeError:
            raise ValueError("%s out of range (%s - %s)" % (value, self.min, self.max))


class RangedInts(RangedInt):
    def __check__(self, values):
        return [super(RangedInts, self).__check__(value) for value in values]


class BoolField(TypedField):
    def __init__(self, value):
        super(BoolField, self).__init__(value, self.to_bool)

    @staticmethod
    def to_bool(val):
        if isinstance(val, str):
            val = val.lower() in ['yes', 'true']
        else:
            val = True if val else False
        return val


class Selection(Structure):
    def __init__(self, *args, **kwargs):
        super(Selection, self).__init__(*args, **kwargs)
        for key in self.__keys__:
            self.__setitem__(key, key)
            setattr(self.__class__, key, key)
        self._readonly = True


class SelectionField(TypedField):
    def __init__(self, value, allowed_values):
        """
        Enforces the value to be one of the allowed values
        :param str|property value:
        :param List[str] | Type[Selection] allowed_values: the list of allowed values
        """
        try:
            if issubclass(allowed_values, Structure):
                allowed_values = allowed_values()
        except TypeError:
            pass
        if isinstance(allowed_values, Structure):
            allowed_values = list(allowed_values.__keys__)

        self.allowed_values = allowed_values
        super(SelectionField, self).__init__(value, self.check)

    def check(self, val):
        if val in self.allowed_values:
            return val
        raise KeyError("'%s' is not an allowed value" % val)

    def __deepcopy__(self, memodict=None):
        n = self.__class__(copy(self.value), copy(self.allowed_values))
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def __getstate__(self):
        return self.value, self.allowed_values


class MultiSelection(SelectionField):
    def __init__(self, values, allowed_values):
        """
        Enforces the values to be in the allowed values
        :param list|tuple values: list of values which all need to be allowed
        :param List[str] | Type[Selection] allowed_values: the list of allowed values
        """
        super(MultiSelection, self).__init__(tuple(values), allowed_values)

    def check(self, vals):
        return tuple((super(MultiSelection, self).check(val) for val in vals))

class Deprecated(object):
    def __init__(self, field=None, new_fieldname=None, migrate=None):
        """
        This is used as a wrapper to denote a field as deprecated.
        When the field is read from yaml it will be run through converter to convert contents
        to new field if needed.
        This field cannot be read by the software end-use any longer
        :param Field field: Original field declaration
        :param str new_fieldname:  name of the new target field the converted output should be stored in
        :param function migrate: function/lambda to migrate value from old field to new field
        """
        if new_fieldname is not None and not isinstance(new_fieldname, str):
            raise ValueError("new_fieldname arg must be the name of the converted target field")

        self.field = field
        self.new_fieldname = new_fieldname
        self.migrate = migrate

        if field is not None:
            self.field_converter = getattr(field, 'converter', None)
            self.field_updater = getattr(field, '_update_field', None)
            self.field_initialiser = getattr(field, "_field_initialiser", lambda: deepcopy(field))

    def __or__(self, other):
        return self

    def __deepcopy__(self, memodict=None):
        n = self.__class__(*deepcopy((self.field, self.new_fieldname, self.migration), memodict or {}))
        n.__doc__ = self.__doc__
        n.__group__ = self.__group__
        return n

    def _update_field(self, previous_value, new_value, parent=None, ignore_errors=False):
        if self.new_fieldname is None:
            return

        if self.field_updater:
            value = self.field_initialiser()
            self.field_updater(value, new_value, parent=parent, ignore_errors=ignore_errors)
            new_value = value
        elif self.field_converter:
            new_value = self.field_converter(new_value)

        if self.migrate:
            new_value = self.migrate(new_value)
        if parent and self.new_fieldname:
            parent[self.new_fieldname] = new_value

    def _field_initialiser(self):
        # Deprecated fields never add a value to output
        return OMIT_VALUE

class Group(object):
    def __init__(self, group_name):
        self.group_name = group_name

    def _field_initialiser(self):
        return OMIT_VALUE

    def _fields_registered(self, fields):
        inside_group = False
        for key, field in fields.items():
            if field == self:
                inside_group = True
            elif inside_group and isinstance(field, Group):
                break
            elif inside_group and hasattr(field, "__group__"):
                setattr(field, "__group__", self.group_name)

